/* Copyright 1999-2004 Red Hat, Inc.
 *
 * This software may be freely redistributed under the terms of the GNU
 * public license.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <ctype.h>
#include <fcntl.h>
#include <ftw.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <errno.h>

#include <sys/mount.h>
#include <sys/poll.h>
#include <sys/stat.h>
#include <sys/utsname.h>

#include <libintl.h>
#include <locale.h>

#include <newt.h>
#include <popt.h>

#define _(String) gettext((String))
#define N_(String) (String)

#include "kudzu.h"

#include "modules.h"

#define FILENAME "hwconf"

static int madebak=0;
static int timeout=0;

struct device **storedDevs;
int numStored=0;
struct device **currentDevs;
int numCurrent=0;

int configuredX = -1;
int removedMouse = 0;

static int rhgb = 0;
static char *_module_file = NULL;
static char *kernel_ver = NULL;

int Xconfig(struct device *dev)
{
	struct device **devlist;
	struct ddcDevice *monitordev = NULL;
	int videomem = 0;
	int x;
	char path[512];
	char tmp[64];
	
	devlist = probeDevices(CLASS_UNSPEC, BUS_DDC, PROBE_ALL);
	if (devlist) {
		for (x = 0; devlist[x]; x++) {
			if (devlist[x]->type == CLASS_MONITOR) 
				monitordev = (struct ddcDevice *)devlist[x];
			else if (devlist[x]->type == CLASS_VIDEO)
				videomem = ((struct ddcDevice *)devlist[x])->mem;
		}
	}
	snprintf(path,352,"/usr/bin/system-config-display --reconfig --noui --set-card=\"%s\"",dev->driver+5);
	if (monitordev && monitordev->horizSyncMin) {
		snprintf(tmp,64," --set-hsync=\"%d-%d\"",monitordev->horizSyncMin,monitordev->horizSyncMax);
		strcat(path,tmp);
	}
	if (monitordev && monitordev->vertRefreshMin) {
		snprintf(tmp,64," --set-vsync=\"%d-%d\"",monitordev->vertRefreshMin,monitordev->vertRefreshMax);
		strcat(path,tmp);
	}
	if (videomem) {
		snprintf(tmp,32," --set-videoram=%d",videomem);
		strcat(path,tmp);
	}
	newtSuspend();
	x=system(path);
	newtResume();
	openlog("kudzu",0,LOG_USER);
	syslog(LOG_NOTICE,"ran system-config-display for %s",dev->driver);
	closelog();
	return x;
}

int makeLink(struct device *dev, char *name)
{
	char oldfname[256],newfname[256];
	
	if (!dev->device || !name) return 1;
	snprintf(oldfname,256,"/dev/%s",dev->device);
	if (dev->index > 0) {
		snprintf(newfname,256,"/dev/%s%d",name,dev->index);
	} else {
		snprintf(newfname,256,"/dev/%s",name);
	}
	openlog("kudzu",0,LOG_USER);
	syslog(LOG_NOTICE,_("linked %s to %s"),newfname,oldfname);
	closelog();
	return symlink(oldfname,newfname);
}

int isLinked(struct device *dev, char *name)
{
	char path[256],path2[256];
	
	memset(path,'\0',256);
	memset(path2,'\0',256);
	if (!name) return 0;
	if (!dev->device) return 0;
	if (dev->index) 
	  snprintf(path,256,"/dev/%s%d",name,dev->index);
	else 
	  snprintf(path,256,"/dev/%s",name);
	if (readlink(path,path2,256)>0) {
		if (!strncmp(path2,"/dev/",5)) {
			if (!strcmp(path2+5,dev->device))
			  return 1;
		} else {
			if (!strcmp(path2,dev->device))
			  return 1;
		}
	}
	return 0;
}

static inline int isSerialish(struct device *dev) {
	return (dev->device && (
			    !strcmp(dev->device, "console") || 
			    !strncmp(dev->device, "ttyS", 4) || 
			    !strcmp(dev->device, "hvc0") ||
			    !strcmp(dev->device, "ttySG0") ||
			    !strncmp(dev->device, "hvsi", 4) ));
}

/* HACK. This is so not thread-safe. */
static char mod_name[100], cmod_name[100]; 

static int isModule(const char *filename, const struct stat *sb, int flag) {
	char *fname = basename(filename);
	/* ugly hack; ignore things under build symlink */
	if ((!strcmp(fname,mod_name) || !strcmp(fname,cmod_name)) && !strstr(filename,"/build/")) {
		return 1;
	}
	return 0;
}


int isAvailable(char *modulename)
{
	struct stat sbuf;
	char path[512];
	int x;
	char *searchdir[] = {
		"fs",
		"net",
		"scsi",
		"block",
		"cdrom",
		"ipv4",
		"ipv6",
		"sound",
		"fc4",
		"video",
		"misc",
		"pcmcia",
		"atm",
		"usb",
		NULL
	};
	
	snprintf(mod_name,100,"%s.ko",modulename);
	snprintf(cmod_name,100,"%s.ko.gz",modulename);
	snprintf(path,512,"/lib/modules/%s",kernel_ver);
	/* Do not set the third argument of this function to < 6. Blarg. */
	if (ftw(path,isModule,15) == 1) {
		return 1;
	}
	/* Finally, try non-version specific directories. */
	for (x = 0; searchdir[x]; x++) {
		snprintf(path,512,"/lib/modules/%s/%s.o",
			 searchdir[x],modulename);
		if (!stat(path,&sbuf))
		  return 1;
		snprintf(path,512,"/lib/modules/%s/%s.o.gz",
			 searchdir[x],modulename);
		if (!stat(path,&sbuf))
		  return 1;
		snprintf(path,512,"/lib/modules/%s/%s.ko",
			 searchdir[x],modulename);
		if (!stat(path,&sbuf))
		  return 1;
	}
	return 0;
}

int isConfigurable(struct device *dev) {
	struct stat tmpstat;
	
	if (!strcmp(dev->driver,"parport_serial"))
		return 1;
	switch (dev->type) {
	 case CLASS_NETWORK:
	 case CLASS_SCSI:
	 case CLASS_IDE:
	 case CLASS_RAID:
	 case CLASS_CAPTURE:
	 case CLASS_USB:
	 case CLASS_FIREWIRE:
	 case CLASS_KEYBOARD:
	 case CLASS_MODEM:
	 case CLASS_SCANNER:
	 case CLASS_PRINTER:
		return 1;
	 case CLASS_MOUSE:
		return 1;
	 case CLASS_AUDIO:
		return 1;
	 case CLASS_VIDEO:
		if (!stat("/usr/bin/system-config-display", &tmpstat) &&
		    (!stat("/usr/X11R6/bin/XFree86", &tmpstat) ||
		     !stat("/usr/X11R6/bin/Xorg", &tmpstat)))
			return 1;
		else
			return 0;
	 case CLASS_OTHER:
		if (dev->bus == BUS_PCI && 
		    ((struct pciDevice *)dev)->vendorId == 0x14e4 &&
		    ((struct pciDevice *)dev)->deviceId == 0x5820)
			return 1;
		return 0;
	 case CLASS_TAPE:
	 case CLASS_FLOPPY:
	 case CLASS_HD:
	 case CLASS_CDROM:
	 case CLASS_SOCKET:
	 default:
		return 0;
	}
	return 0;
}

int isConfigured(struct device *dev)
{
	struct confModules *cf;
	char path[256], path2[256];
	struct stat sbuf;
	int ret=0;
	
	memset(path,'\0',256);
	memset(path2,'\0',256);
	cf = readConfModules(_module_file);
	switch (dev->type) {
	 case CLASS_NETWORK:
		if (!strcmp(dev->driver,"unknown") ||
		    !strcmp(dev->driver,"ignore") ||
		    !strcmp(dev->driver,"disabled"))
		  ret=1;
		/* Assume cardbus stuff is unconfigured */
		if (dev->bus == BUS_PCMCIA)
			break;
		if (dev->bus != BUS_PCI || ((struct pciDevice *)dev)->pciType != PCI_CARDBUS)
		  if (cf)
		    if (dev->device && isAliased(cf,dev->device,dev->driver)!=-1)
		      ret = 1;
		break;
	 case CLASS_RAID:
	 case CLASS_SCSI:
	 case CLASS_IDE:
		if (!strcmp(dev->driver,"unknown") ||
		    !strcmp(dev->driver,"disabled") ||
		    !strcmp(dev->driver,"ignore"))
		  ret=1;
		if (cf)
		  if (isAliased(cf,"scsi_hostadapter",dev->driver)!=-1)
		    ret=1;
		break;
	 case CLASS_VIDEO:
	 case CLASS_MONITOR:
		/* Assume on initial runs that if X is configured, we got the right card */
		ret = 1;
		if (!strcmp(dev->driver,"unknown") ||
		    !strcmp(dev->driver,"disabled") ||
		    !strcmp(dev->driver,"ignore"))
		  ret=1;
		break;
	 case CLASS_CAPTURE:
		if (!strcmp(dev->driver,"unknown") ||
		    !strcmp(dev->driver,"disabled") ||
		    !strcmp(dev->driver,"ignore"))
		  ret=1;
		if (cf)
		  if (isAliased(cf,"char-major-81",dev->driver)!=-1)
		    ret = 1;
		break;
	 case CLASS_AUDIO:
		if (!strcmp(dev->driver,"unknown") ||
		    !strcmp(dev->driver,"disabled") ||
		    !strcmp(dev->driver,"ignore"))
		  ret=1;
		if (cf) {
			if (isAliased(cf,"sound",dev->driver)!=-1)
			  ret = 1;
			if (isAliased(cf,"sound-card-",dev->driver)!=-1)
			  ret = 1;
			if (isAliased(cf,"sound-slot-",dev->driver)!=-1)
			  ret = 1;
			if (isAliased(cf,"snd-card-",dev->driver)!=-1)
			  ret = 1;
		}
		break;
	 case CLASS_USB:
	        if (!strcmp(dev->driver,"unknown") ||
		    !strcmp(dev->driver,"disabled") ||
		    !strcmp(dev->driver,"ignore"))
		    ret = 1;
		if (cf && isAliased(cf,"usb-controller",dev->driver)!=-1)
		  ret = 1;
		break;
	 case CLASS_FIREWIRE:
	        if (!strcmp(dev->driver,"unknown") ||
		    !strcmp(dev->driver,"disabled") ||
		    !strcmp(dev->driver,"ignore"))
		    ret = 1;
		if (cf && isAliased(cf,"ieee1394-controller",dev->driver)!=-1)
		  ret = 1;
		break;
	 case CLASS_MOUSE:
		{
			int fd;
			char *buf, *tmp, devpath[256];

			fd = open("/etc/sysconfig/mouse",O_RDONLY);
			if (fd < 0) {
				ret = 0;
				break;
			}
			buf = bufFromFd(fd);
			tmp = strstr(buf,"DEVICE=");
			if (!tmp) {
				ret = 0;
				break;
			}
			if (tmp == buf || *(tmp-1) == '\n') {
				snprintf(devpath,255,"/dev/%s",dev->device);
				if (!strncmp(tmp+7,devpath, strlen(devpath))) {
					ret = 1;
					break;
				}
			}
			ret = 0;
		}
		break;
			
	 case CLASS_MODEM:
		ret = isLinked(dev,"modem");
		break;
	 case CLASS_SCANNER:
		ret = isLinked(dev,"scanner");
		break;
	 case CLASS_KEYBOARD:
		break;	 
	 case CLASS_PRINTER:
		/* fairly braindead printcap parser */
		{
			char *buf,*ptr,*tmp;
			int fd;
			
			fd = open("/etc/printcap",O_RDONLY);
			if (fd==-1) break;
			fstat(fd,&sbuf);
			buf=malloc(sbuf.st_size+1);
			if (read(fd,buf,sbuf.st_size)!=sbuf.st_size) break;
			buf[sbuf.st_size] = '\0';
			ptr=buf;
			while (buf[0]!='\0') {
				if (ptr[0] == '#') {
					while (*ptr && *ptr != '\n') ptr++;
					if (*ptr) {
						*ptr='\0';
						ptr++;
					}
					buf=ptr;
		                        continue;
				}
				while (*ptr && *ptr != '\0' && *ptr !='\n') ptr++;
				if (*ptr) {
					*ptr='\0';
					ptr++;
				}
				if ((tmp=strstr(buf,"lp=/dev/"))) {
					while (*tmp && *tmp != ':') tmp++;
					if (*tmp) {
						*tmp = '\0';
					}
					if (!strcmp(buf+8,dev->device))
					  ret = 1;
				}
				buf = ptr;
			}
			break;
		}
	 case CLASS_OTHER:
		if (dev->bus == BUS_PCI && 
		    ((struct pciDevice *)dev)->vendorId == 0x14e4 &&
		    ((struct pciDevice *)dev)->deviceId == 0x5820)
			return 0;
		return 1;
	 default:
		/* If we don't know how to configure it, assume it's configured. */
		ret = 1;
		break;
	}
	if (cf)
	  freeConfModules(cf);
	return ret;
}

int configure(struct device *dev)
{
	struct confModules *cf;
	char path[256];
	int index;
	static int current_netdev = 0;

	if (!strcmp(dev->driver,"parport_serial")) {
		cf = readConfModules(_module_file);
		if (!cf)
		  cf = newConfModules();
		cf->madebackup = madebak;
		addAlias(cf,"parport_lowlevel","parport_serial", CM_REPLACE);
		writeConfModules(cf,_module_file);
		madebak = cf->madebackup;
		openlog("kudzu",0,LOG_USER);
		syslog(LOG_NOTICE,_("aliased %s as %s"),"parport_lowlevel","parport_serial");
		closelog();
	} else switch (dev->type) {
	 case CLASS_NETWORK:
		if (!dev->device) break;
		cf = readConfModules(_module_file);
		if (!cf) 
		  cf = newConfModules();
		cf->madebackup = madebak;
		if (isAliased(cf,dev->device,dev->driver)) {
			if ( (dev->bus != BUS_PCI && dev->bus != BUS_PCMCIA) || 
			     ((struct pciDevice *)dev)->pciType != PCI_CARDBUS) {
				addAlias(cf,dev->device,dev->driver,CM_REPLACE);
				writeConfModules(cf,_module_file);
				madebak = cf->madebackup;
				openlog("kudzu",0,LOG_USER);
				syslog(LOG_NOTICE,_("aliased %s as %s"),path,dev->driver);
				closelog();
			}
		}
		freeConfModules(cf);
		current_netdev++;
		break;
	 case CLASS_RAID:
	 case CLASS_IDE:
	 case CLASS_SCSI:
		cf = readConfModules(_module_file);
		if (!cf)
		  cf = newConfModules();
		cf->madebackup = madebak;
		if (isAliased(cf,"scsi_hostadapter",dev->driver)==-1) {
			index=0;
			while (1) {
				if (index)
				  snprintf(path,256,"scsi_hostadapter%d",index);
				else
				  snprintf(path,256,"scsi_hostadapter");
				if (getAlias(cf,path)) 
				  index++;
				else
				  break;
			}
			addAlias(cf,path,dev->driver,CM_REPLACE);
			writeConfModules(cf,_module_file);
			madebak = cf->madebackup;
			openlog("kudzu",0,LOG_USER);
			syslog(LOG_NOTICE,_("aliased %s as %s"),path,dev->driver);
			closelog();
		}
		freeConfModules(cf);
		break;
	 case CLASS_VIDEO:
		configuredX = Xconfig(dev);
		break;
	 case CLASS_CAPTURE:
		cf = readConfModules(_module_file);
		if (!cf)
		  cf = newConfModules();
		cf->madebackup = madebak;
		if (isAliased(cf,"char-major-81",dev->driver)==-1) {
			snprintf(path,256,"char-major-81");
			addAlias(cf,path,dev->driver,CM_REPLACE);
			writeConfModules(cf,_module_file);
			madebak = cf->madebackup;
			openlog("kudzu",0,LOG_USER);
			syslog(LOG_NOTICE,_("aliased %s as %s"),path,dev->driver);
			closelog();
		}
		freeConfModules(cf);
		break;
	 case CLASS_USB:
		cf = readConfModules(_module_file);
		if (!cf)
		  cf = newConfModules();
		cf->madebackup = madebak;
		if (isAliased(cf,"usb-controller",dev->driver)==-1) {
			index=0;
			while (1) {
				if (index)
				  snprintf(path,256,"usb-controller%d",index);
				else
				  snprintf(path,256,"usb-controller");
				if (getAlias(cf,path))
				  index++;
				else
				  break;
			}
			addAlias(cf,path,dev->driver,CM_REPLACE);
			writeConfModules(cf,_module_file);
			madebak = cf->madebackup;
			openlog("kudzu",0,LOG_USER);
			syslog(LOG_NOTICE,_("aliased %s as %s"),path,dev->driver);
			closelog();
		}
		freeConfModules(cf);
		break;
	 case CLASS_FIREWIRE:
		cf = readConfModules(_module_file);
		if (!cf)
		  cf = newConfModules();
		cf->madebackup = madebak;
		if (isAliased(cf,"ieee1394-controller",dev->driver)==-1) {
			index=0;
			while (1) {
				if (index)
				  snprintf(path,256,"ieee1394-controller%d",index);
				else
				  snprintf(path,256,"ieee1394-controller");
				if (getAlias(cf,path))
				  index++;
				else
				  break;
			}
			addAlias(cf,path,dev->driver,CM_REPLACE);
			writeConfModules(cf,_module_file);
			madebak = cf->madebackup;
			openlog("kudzu",0,LOG_USER);
			syslog(LOG_NOTICE,_("aliased %s as %s"),path,dev->driver);
			closelog();
		}
		freeConfModules(cf);
		break;
	 case CLASS_AUDIO:
		cf = readConfModules(_module_file);
		if (!cf)
			cf = newConfModules();
		cf->madebackup = madebak;
		if (
		    isAliased(cf,"sound-slot-",dev->driver)==-1 &&
		    isAliased(cf,"sound-card-",dev->driver)==-1 &&
		    isAliased(cf,"snd-card-",dev->driver)==-1
		    ) {
			snprintf(path,256,"snd-card-%d",dev->index);
			addAlias(cf,path,dev->driver,CM_REPLACE);
			snprintf(path,256,"options snd-card-%d index=%d",dev->index,dev->index);
			addLine(cf,path,CM_REPLACE);
			snprintf(path,256,"install %s /sbin/modprobe --ignore-install %s && /usr/sbin/alsactl restore >/dev/null 2>&1 || :", dev->driver, dev->driver);
			addLine(cf,path,CM_REPLACE);
			snprintf(path,256,"remove %s { /usr/sbin/alsactl store >/dev/null 2>&1 || : ; }; /sbin/modprobe -r --ignore-remove %s",dev->driver, dev->driver);
			addLine(cf,path,CM_REPLACE);
			writeConfModules(cf,_module_file);
			madebak = cf->madebackup;
			openlog("kudzu",0,LOG_USER);
			snprintf(path,256,"snd-card-%d",dev->index);
			syslog(LOG_NOTICE,_("aliased %s as %s"),path,dev->driver);
			closelog();
		}
		break;
	 case CLASS_MOUSE:
		if (!strcmp(dev->driver,"ignore")) break;
		snprintf(path,256,"/usr/sbin/mouseconfig --kickstart --device %s %s",
			 dev->device, dev->driver);
		system(path);
		removedMouse = 0;
		openlog("kudzu",0,LOG_USER);
		syslog(LOG_NOTICE,_("ran mouseconfig for %s"),dev->device);
		closelog();
		break;
	 case CLASS_KEYBOARD:
		break;
	 case CLASS_MODEM:
		makeLink(dev,"modem");
		break;
	 case CLASS_SCANNER:
		makeLink(dev,"scanner");
		break;
	 case CLASS_PRINTER:
		newtSuspend();
		if (dev->bus == BUS_PARALLEL) {
			snprintf(path,256,"/usr/sbin/printconf-tui --Xadd-local --device=/dev/%s --make=\"%s\" --model=\"%s\" >/dev/null 2>&1",
				 dev->device,
				 ((struct parallelDevice *)dev)->pnpmfr,
				 ((struct parallelDevice *)dev)->pnpmodel);
		} else if (dev->bus == BUS_USB) {
			if (((struct usbDevice *)dev)->usbmfr &&
			    ((struct usbDevice *)dev)->usbprod)
				snprintf(path,256,"/usr/sbin/printconf-tui --Xadd-local --device=/dev/usb/lp0 --make=\"%s\" --model=\"%s\" >/dev/null 2>&1",
				 ((struct usbDevice *)dev)->usbmfr,
				 ((struct usbDevice *)dev)->usbprod);
			else break;
		} else
			break;
		system(path);
		newtResume();
		openlog("kudzu",0,LOG_USER);
		syslog(LOG_NOTICE,_("ran printconf for %s"),dev->desc);
		closelog();
		break;
	 case CLASS_OTHER:
		if (dev->bus == BUS_PCI && 
		    !strcmp(dev->driver,"bcm5820")) {
			int retcode;
			
			retcode = system("/sbin/chkconfig --level 345 bcm5820 on > /dev/null 2>&1");
			if (retcode == 0) {
				openlog("kudzu",0,LOG_USER);
				syslog(LOG_NOTICE,_("turned on bcm5820 service"));
				closelog();
			}
		}
		if (dev->bus == BUS_PCI && 
		    !strcmp(dev->driver,"paep")) {
			int retcode;
			
			retcode = system("/sbin/chkconfig --level 345 aep1000 on > /dev/null 2>&1");
			if (retcode == 0) {
				openlog("kudzu",0,LOG_USER);
				syslog(LOG_NOTICE,_("turned on aep1000 service"));
				closelog();
			}
		}
		break;
	 case CLASS_TAPE:
	 case CLASS_FLOPPY:
	 case CLASS_HD:
	 default:
		break;
	}
	return 0;
}

void configMenu(struct device *oldDevs, struct device *newDevs)
{
	int y, z, rc;
	struct device *dev, *tmpdev;
	int defaction = -1;
	int mouseconfigured = 0;
	
	dev = newDevs;
	for ( ; dev; dev=dev->next) {
		if (isConfigurable(dev) &&
		    !(dev->bus == BUS_PCI &&  dev->type != CLASS_MODEM &&
		      (!strcmp(dev->driver, "ignore") || 
		       !strcmp(dev->driver, "disabled") ||
		       !strcmp(dev->driver, "unknown"))) ) {
		  if (!isConfigured(dev)) {
			  switch (dev->type) {
			   case CLASS_NETWORK:
			   case CLASS_SCSI:
			   case CLASS_IDE:
			   case CLASS_RAID:
			   case CLASS_CAPTURE:
			   case CLASS_AUDIO:
			   case CLASS_USB:
			   case CLASS_FIREWIRE:
			   case CLASS_OTHER:
				  if (isAvailable(dev->driver))
				    continue;
				  break;
			   /* If we are running for the first time, and they
			    * have a mouse that currently exists configured,
			    * ignore any secondary mice. */
			   case CLASS_MOUSE:
				  if (mouseconfigured)
				    break;
				  continue;
			   default:
				  continue;
			  }

		  }
		}
		newDevs = listRemove(newDevs, dev);
	}
	if (!newDevs)
	  return;
	
	dev = newDevs;
	for ( ; dev ; dev = dev->next) {
		rc = defaction >= 0 ? defaction : 0;
		switch (rc) {
		 case 0:
		 case 3:
			if (rc == 3) defaction = 0;
			configure(dev);
			break;
		 case 1:
		 case 4:
			if (rc == 4) defaction = 1;
			break;
		 case 2:
			y=0;
			while (currentDevs[y]) {
				if (currentDevs[y]==dev) {
					for (z=y;z<numCurrent;z++)
					  currentDevs[z]=currentDevs[z+1];
					numCurrent--;
					break;
				}
				y++;
			}
			break;
		 case 5:
			newtFinished();
			exit(0);
			break;
		}
	}
	if (configuredX >= 0) {
		if (configuredX) {
			close(open("/var/run/Xconfig-failed",O_CREAT|O_EXCL,0644));
		} else {
			close(open("/var/run/Xconfig",O_CREAT|O_EXCL,0644));
		}
	}
	if (removedMouse) {
		int i;
		int fixedmouse = 0;
		
		for (i=0; currentDevs[i]; i++) {
			if (currentDevs[i]->type == CLASS_MOUSE &&
			    strcmp(currentDevs[i]->driver,"ignore")) {
				configure(currentDevs[i]);
				fixedmouse = 1;
				break;
			}
		}
		if (!fixedmouse) {
			close(open("/var/run/Xconfig-failed",O_CREAT|O_EXCL,0644));
		}
	}
}

static void setupKernelVersion() {
        unsigned int major, sub, minor;

        if (!kernel_ver) {
                struct utsname utsbuf;

                uname(&utsbuf);
                kernel_ver = strdup(utsbuf.release);
        }
}

int main(int argc, char **argv) 
{
	char *confFile;
	int ret;
	int rc;
	int x;
	enum deviceBus probeBus = BUS_UNSPEC;
	enum deviceClass probeClass = CLASS_UNSPEC;
	struct device **oldDevs, **newDevs;

	if (getuid()) {
		fprintf(stderr,
			_("\nERROR - You must be root to run kudzu.\n"));
		exit(1);
	}
	
	setupKernelVersion();
	_module_file = "/etc/modprobe.conf";
	
	initializeBusDeviceList(probeBus);
	storedDevs = malloc(sizeof(struct device *));
	storedDevs[0] = NULL;
			       
	while (storedDevs[numStored]) numStored++;
	currentDevs = readDevices("/etc/sysconfig/hwconf");
	if (!currentDevs) {
		currentDevs = malloc(sizeof(struct device *));
		currentDevs[0] = NULL;
	}
	while (currentDevs[numCurrent]) numCurrent++;
	ret = listCompare(storedDevs, currentDevs, &oldDevs, &newDevs);
	freeDeviceList();
	if (!ret) {
		exit(0);
	} else {
		/* List-ify oldDevs, newDevs */
		if (oldDevs[0]) {
			oldDevs[0]->next=NULL;
			for (x=1;oldDevs[x];x++)
				oldDevs[x-1]->next = oldDevs[x];
			oldDevs[x-1]->next = NULL;
		}
		if (newDevs[0]) {
			newDevs[0]->next = NULL;
			for (x=1;newDevs[x];x++) 
				newDevs[x-1]->next = newDevs[x];
			newDevs[x-1]->next = NULL;
		}
	        configMenu((*oldDevs),(*newDevs));
	}
	
	newtFinished();
	if (!access("/usr/bin/rhgb-client",X_OK) && rhgb) {
		system("/usr/bin/rhgb-client --details=no >/dev/null 2>&1");
	}
	return 0;
}
